unit context;
	{ This unit deals with text output & keyboard input. }
{
	GearHead: Arena, a roguelike mecha CRPG
	Copyright (C) 2005 Joseph Hewitt

	This library is free software; you can redistribute it and/or modify it
	under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation; either version 2.1 of the License, or (at
	your option) any later version.

	The full text of the LGPL can be found in license.txt.

	This library is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
	General Public License for more details. 

	You should have received a copy of the GNU Lesser General Public License
	along with this library; if not, write to the Free Software Foundation,
	Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
}

interface

uses
{$IF DEFINED(DEBUG) or DEFINED(PATCH_I18N) or DEFINED(PATCH_GH) or DEFINED(PATCH_HEART)}
	sysutils,
{$ENDIF}
{$IFDEF ASCII_SDL}
  {$IFDEF SDL2}
	SDL2,
  {$ELSE SDL2}
	SDL,
  {$ENDIF SDL2}
{$ENDIF ASCII_SDL}
{$IFDEF PATCH_GH}
	gears_base,
{$ENDIF PATCH_GH}
	gears;

const
	Console_History_Length = 240;

var
	Text_Messages: SAttPtr;
	Console_History: SAttPtr;

{$IFDEF PATCH_I18N}
Procedure WriteMBCharStr( arg_msg: String; Xwidth: Integer );
{$ENDIF PATCH_I18N}

Function RPGKey: Char;
{$IFDEF PATCH_HEART}
Function RPGKey_nonblocking: Char;
{$ENDIF PATCH_HEART}
Function DirKey: Integer;
Procedure EndOfGameMoreKey;
{$IFDEF ASCII_SDL}
Procedure CMessage( const msg: String; Z: Integer; C: TSDL_Color );
Procedure GameMSG( msg: string; X1,Y1,X2,Y2: Byte; C: TSDL_Color ); {no const}
Procedure GameMSG( const msg: string; Z: Byte; C: TSDL_Color );
{$ELSE ASCII_SDL}
Procedure CMessage( const msg: String; Z: Integer; C: Byte );
Procedure GameMSG( msg: string; X1,Y1,X2,Y2,C: Byte ); {no const}
Procedure GameMSG( const msg: string; Z,C: Byte );
{$ENDIF ASCII_SDL}
{$IFDEF PATCH_GH}
Procedure RedrawConsole;
{$ENDIF PATCH_GH}
Procedure DialogMSG(msg: string); {no const}
{$IFDEF PATCH_GH}
Function GetStringFromUser( const Prompt, Init_text: String; var canceled: Boolean ): String;
Function GetStringFromUser( const Prompt, Init_text: String ): String;
{$ENDIF PATCH_GH}
Function GetStringFromUser( const Prompt: String ): String;
Function MsgString( const MsgLabel: String ): String;

{$IFDEF PATCH_GH}
{$ELSE PATCH_GH}
Function MoreHighFirstLine( LList: SAttPtr ): Integer;
{$ENDIF PATCH_GH}
Procedure MoreText( LList: SAttPtr; FirstLine: Integer );

{$IFDEF PATCH_GH}
	{ SetupHQDisplay was moved from congfx.pp. }
Procedure SetupHQDisplay;
{$ELSE PATCH_GH}
{$ENDIF PATCH_GH}

implementation

uses
{$IFDEF PATCH_I18N}
  {$IFDEF WITH_WIDECHAR}
	strings,
  {$ELSE WITH_WIDECHAR}
    {$IFDEF GUIMSWINMODE}
	strings,
    {$ENDIF GUIMSWINMODE}
  {$ENDIF WITH_WIDECHAR}
{$ENDIF PATCH_I18N}
{$IF DEFINED(ASCII_SDL)}
{$ELSEIF DEFINED(GUIMSWINMODE)}
{$ELSE}
	crt,
{$ENDIF}
{$IFDEF DEBUG}
	errmsg,
{$ENDIF DEBUG}
{$IFDEF PATCH_GH}
	pseudosmartpointer,
{$ENDIF PATCH_GH}
{$IFDEF PATCH_I18N}
  {$IFDEF PATCH_GH}
	i18nmsg,
  {$ENDIF PATCH_GH}
  {$IFDEF WITH_TENC}
	termenc,
  {$ENDIF WITH_TENC}
  {$IF DEFINED(ASCII_SDL)}
	ascii_sdl,
  {$ELSEIF DEFINED(GUIMSWINMODE)}
	w32crt,
  {$ENDIF}
{$ENDIF PATCH_I18N}
	texutil,ui4gh,congfx
{$IFDEF PATCH_GH}
	,conmenus
{$ENDIF PATCH_GH}
	;

{$IFDEF PATCH_GH}
Procedure GetNextLine( var TheLine , msg , NextWord: String; Width: Integer );
	{ Get a line of text of maximum width "Width". }
var
	LC: Boolean;	{ Loop Condition. So I wasn't very creative when I named it, so what? }
	CW_I18N: Boolean;	{Is the current word I18N ?}
	DItS: Boolean;		{Do insert the space, or not.}
	BW: String;
begin
	{ Loop condition starts out as TRUE. }
	LC := True;
	if ( ' ' = TheLine ) then TheLine := '';

	{ Start building the line. }
	repeat
		NextWord := ExtractWordForPrint( Msg, DItS, CW_I18N );

		if ( '' <> NextWord ) then begin
			if ( False = CW_I18N ) then begin
				if ( WidthMBCharStr( TheLine + ' ' + NextWord) < Width ) then
					if DItS then TheLine := TheLine + ' ' + NextWord
					else TheLine := TheLine + NextWord
				else
					LC := False;
			end else begin
				if ( WidthMBCharStr( TheLine + NextWord + ' ' + ' ') < Width ) then
					if DItS then TheLine := TheLine + ' ' + NextWord
					else TheLine := TheLine + NextWord
				else begin
					LC := False;

					if ( 0 < Pos(NextWord, ProhibitationHead) ) then begin
						TheLine := TheLine + NextWord + #13;
					end else begin
						BW := TailMBChar(TheLine);
						if (0 < Length(BW)) and (0 < Pos(BW, ProhibitationTrail)) then begin
							TheLine := Copy(TheLine,1,Length(TheLine)-Length(BW));
							NextWord := BW + NextWord;
						end;
					end;
				end;
			end;
		end;
	until (not LC) or (NextWord = '') or ( (0 < Length(TheLine)) and (TheLine[Length(TheLine)] = #13) );

	{ If the line ended due to a line break, deal with it. }
	if ( (0 < Length(TheLine)) and (TheLine[Length(TheLine)] = #13) ) then begin
		{ Display the line break as a space. }
		TheLine[Length(TheLine)] := ' ';
		NextWord := ExtractWordForPrint( Msg, DItS, CW_I18N );
	end;
end;
{$ENDIF PATCH_GH}

{$IFDEF PATCH_I18N}
Procedure WriteMBCharStr( arg_msg: String; Xwidth: Integer );
	{ NOTE: In CJK, there are many charctors, one charactor have double size for one ANK charactor and data length is 3 or 4 bytes. }
	{ NOTE: But, a function Write() clipped by data length. }
	{ NOTE: FPC's unicode functions is not stable, is its? }
  {$IF DEFINED(ASCII_SDL)}
  {$ELSEIF DEFINED(GUIMSWINMODE)}
const
	WCLen = 288; { 254; }
  {$ELSE}
    {$IFDEF WITH_WIDECHAR}
const
	WCLen = 512;
    {$ELSE WITH_WIDECHAR}
      {$IFDEF WITH_TENC}
const
	tmsgLen = 288; { 254; }
      {$ENDIF WITH_TENC}
    {$ENDIF WITH_WIDECHAR}
  {$ENDIF}

  {$IF DEFINED(ASCII_SDL)}
  {$ELSEIF DEFINED(GUIMSWINMODE)}
var
	pmsg: PChar;
	pdst: PChar;
	dstlen: Integer;
	dst: String;
  {$ELSE}
var
    {$IFDEF WITH_WIDECHAR}
	pmsg: PChar;
	pdst: PWideChar;
	dstlen: Integer;
	wdst: WideString;
    {$ELSE WITH_WIDECHAR}
	msg: String;
      {$IFDEF WITH_TENC}
	tmsg: Array[0..tmsgLen] of Char;
	pmsg, ptmsg: PChar;
      {$ENDIF WITH_TENC}
	MaxLen: Integer;
	P, lastP: Integer;
	X, Y: Integer;
	Len: Integer;
    {$ENDIF WITH_WIDECHAR}
  {$ENDIF}

begin
	if (1 < Length(arg_msg)) and (#$0 = arg_msg[1]) then begin
		arg_msg := Copy( arg_msg , 2 , ( Length(arg_msg) - 1 ) );
	end;
  {$IF DEFINED(ASCII_SDL)}
	ascii_sdl.Write( arg_msg );
  {$ELSEIF DEFINED(GUIMSWINMODE)}
	pmsg := QuickPCopy(arg_msg);
	pdst := sysutils.StrAlloc(WCLen);
	dstlen := Conv_ToTenc( pmsg, Length(arg_msg), pdst, WCLen );
	dst := pdst;
	w32crt.Write( dst );
	sysutils.StrDispose( pdst );
	sysutils.StrDispose( pmsg );
  {$ELSE}
    {$IFDEF WITH_WIDECHAR}
	pmsg := QuickPCopy(arg_msg);
	pdst := PWideChar(sysutils.StrAlloc(WCLen));
	dstlen := Conv_ToUni16( pmsg, Length(arg_msg), PWord(pdst), WCLen );
	wdst := pdst;
	Write( wdst );
	sysutils.StrDispose( pdst );
	sysutils.StrDispose( pmsg );
    {$ELSE WITH_WIDECHAR}
      {$IFDEF WITH_TENC}
	if SYSTEM_CHARSET = TERMINAL_CHARSET then begin
		if TERMINAL_bidiRTL then begin
			msg := Conv_bidiRTL(arg_msg);
		end else begin
			msg := arg_msg;
		end;
	end else begin
		ptmsg := tmsg;
		if TERMINAL_bidiRTL then begin
			pmsg := QuickPCopy(Conv_bidiRTL(arg_msg));
		end else begin
			pmsg := QuickPCopy(arg_msg);
		end;
		Conv_ToTenc( pmsg, Length(arg_msg), ptmsg, tmsgLen );
		sysutils.StrDispose( pmsg );
		msg := StrPas( tmsg );
	end;
      {$ELSE WITH_TENC}
	msg := arg_msg;
      {$ENDIF WITH_TENC}

	MaxLen := Length(msg);
	P := 1;	lastP := 1;
	X := WhereX;	Y := WhereY;
	if TERMINAL_bidiRTL and (0 < Xwidth) then begin
		X := X + Xwidth - WidthMBCharStr(msg);
		if X < GOTOXY_MIN then begin
			X := GOTOXY_MIN;
		end;
		GotoXY(X,Y);
	end;

      {$IFDEF PASCAL_WRITE_BUG_HACK}
	while (P <= MaxLen) do begin
        {$IFDEF WITH_TENC}
		Len := LengthMBChar( msg[P], TENC );
        {$ELSE WITH_TENC}
		Len := LengthMBChar( msg[P] );
        {$ENDIF WITH_TENC}
		if 0 < Len then begin
			if Len <= 2 then begin
				{ BUG: Display width is mistook if use JISx0201-KANA. }
				P := P + Len;
				X := X + Len;
			end else begin
				Write(Copy(msg,lastP,P-lastP+len));
				P := P + Len;
				lastP := P;
				X := X + 2;
				GotoXY(X,Y+1);
				if (Y+1) <> WhereY then begin
					GotoXY(X,Y-1);
					if (Y-1) <> WhereY then begin
						GotoXY(X+2,Y);
					end;
				end;
				GotoXY(X,Y);
			end;
		end else begin
			Inc(P);
			Inc(X);
		end;
	end;
	if lastP < P then Write(Copy(msg,lastP,MaxLen-lastP+1));
      {$ELSE PASCAL_WRITE_BUG_HACK}
	Write( msg );
      {$ENDIF PASCAL_WRITE_BUG_HACK}
    {$ENDIF WITH_WIDECHAR}
  {$ENDIF}
end;
{$ENDIF PATCH_I18N}


{$IFDEF ASCII_SDL}
Function RPGKey: Char;
begin
	RPGKey := RPGKey_Raw;
end;
{$ELSE ASCII_SDL}
Function RPGKey: Char;
	{Read a keypress from the keyboard. Convert it into a form}
	{that my other procedures would be willing to call useful.}
var
	rk,getit: Char;
  {$IFDEF PATCH_GH}
	i: Integer;
  {$ENDIF PATCH_GH}
begin
  {$IFDEF GUIMSWINMODE}
	w32crt.RedrawScreen;
	RK := w32crt.ReadKey;
  {$ELSE GUIMSWINMODE}
	RK := ReadKey;
  {$ENDIF GUIMSWINMODE}
  {$IFDEF DEBUG}
	WriteLn(IntToHex(Ord(RK),2));
  {$ENDIF DEBUG}

	Case RK of
		{ 45 5B 1B: 5 in TenKey on Unix-Console }
		#0: begin	{We have a two-part special key.}
			{Obtain the scan code.}
  {$IFDEF GUIMSWINMODE}
			getit := w32crt.ReadKey;
  {$ELSE GUIMSWINMODE}
			getit := ReadKey;
  {$ENDIF GUIMSWINMODE}
  {$IFDEF DEBUG}
			WriteLn(IntToHex(Ord(getit),2));
  {$ENDIF DEBUG}
			case getit of
  {$IFDEF PATCH_GH}
				#$52: RK := '0';	{ 0 in TenKey }
				#$53: RK := '.';	{ . in TenKey on X-Window/MS-Windows }
				#$5A: RK := KeyMap[ KMC_SouthWest ].KCode;	{ End Cursor Key on X-Window }
				#$4F: RK := KeyMap[ KMC_SouthWest ].KCode;	{ End Cursor Key on Unix-Console, MS-Windows }
				#$50: RK := KeyMap[ KMC_South ].KCode;		{ Down Cursor Key }
				#$51: RK := KeyMap[ KMC_SouthEast ].KCode;	{ PageDown Cursor Key }
				#$4B: RK := KeyMap[ KMC_West ].KCode;		{ Left Cursor Key }
				#$5F: RK := '5';				{ 5 in TenKey on X-Window }
				#$4C: RK := '5';				{ 5 in TenKey on MS-Windows }
				#$4D: RK := KeyMap[ KMC_East ].KCode;		{ Right Cursor Key }
				#$5B: begin
						getit := ReadKey;
						case getit of
							#$1B: RK := KeyMap[ KMC_NorthWest ].KCode;	{ Home Cursor Key on X-Window }
							else RK := #0;
						end;
					end;
				#$47: RK := KeyMap[ KMC_NorthWest ].KCode;	{ Home Cursor Key on Unix-Console, MS-Windows }
				#$48: RK := KeyMap[ KMC_North ].KCode;		{ Up Cursor Key }
				#$49: RK := KeyMap[ KMC_NorthEast ].KCode;	{ PageUp Cursor Key }
				#$35: RK := '/';	{ Slash in TenKey on MS-Windows }
				#$1C: RK := ' ';	{ Enter in TenKey on MS-Windows => Altanative-RET }
				{ #$3B: }{ F1  on Unix-Console, MS-Windows }
				{ #$3C: }{ F2  on Unix-Console, MS-Windows }
				{ #$3D: }{ F3  on Unix-Console, MS-Windows }
				{ #$3E: }{ F4  on Unix-Console, MS-Windows }
				{ #$3F: }{ F5  on Unix-Console, MS-Windows }
				{ #$40: }{ F6  }
				{ #$41: }{ F7  }
				{ #$42: }{ F8  }
				{ #$43: }{ F9  }
				{ #$44: }{ F10 }
				{ #$85: }{ F11 }
				{ #$86: }{ F12 }
				#$29: RK := #$1B;	{ JP109key-Hankaku/Zenkaku on MS-Windows => ESC }
  {$ELSE PATCH_GH}
				#72: RK := KeyMap[ KMC_North ].KCode; {Up Cursor Key}
				#71: RK := KeyMap[ KMC_NorthWest ].KCode; {Home Cursor Key}
				#73: RK := KeyMap[ KMC_NorthEast ].KCode; {PageUp Cursor Key}
				#80: RK := KeyMap[ KMC_South ].KCode; {Down Cursor Key}
				#79: RK := KeyMap[ KMC_SouthWest ].KCode; {End Cursor Key}
				#81: RK := KeyMap[ KMC_SouthEast ].KCode; {PageDown Cursor Key}
				#75: RK := KeyMap[ KMC_West ].KCode; {Left Cursor Key}
				#77: RK := KeyMap[ KMC_East ].KCode; {Right Cursor Key}
  {$ENDIF PATCH_GH}
  {$IFDEF PATCH_CHEAT}
				#$86: if Cheat_Display then if Cheat_Display_SW then Cheat_Display_SW := False else Cheat_Display_SW := True;
  {$ENDIF PATCH_CHEAT}
  {$IFDEF PATCH_GH}
				else RK := #0;
  {$ENDIF PATCH_GH}
			end;
		end;

		{Convert the Backspace character to ESCape.}
		#8: RK := #27;	{ Backspace, DEL in TenKey on Unix-Console => ESC }

		{Normally, SPACE is the selection button, but ENTER should}
		{work as well. Therefore, convert all enter codes to spaces.}
		#10: RK := ' ';
		#13: RK := ' ';	{ Enter, Enter in TenKey on Unix-Console/X-Window }
	end;

  {$IFDEF PATCH_GH}
	i := 1;
	while ( KBD_ERROR <> KCodeAlias[i].KCode_from ) do begin
		if ( KCodeAlias[i].KCode_from = RK ) then begin
			RK := KCodeAlias[i].KCode_to;
			break;
		end;
		Inc( i );
	end;
  {$ENDIF PATCH_GH}

	RPGKey := RK;
end;
{$ENDIF ASCII_SDL}

{$IFDEF PATCH_HEART}
  {$IFDEF ASCII_SDL}
Function RPGKey_nonblocking: Char;
var
    {$IFDEF SDL2}
	dummy: String;
    {$ELSE SDL2}
	dummy: Word;
    {$ENDIF SDL2}
begin
	RPGKey_nonblocking := RPGKey_Raw( True , dummy , False );
end;
  {$ELSE ASCII_SDL}
Function RPGKey_nonblocking: Char;
var
	kp: Boolean;
	a: Char;
	t: LongInt;
begin
	a := #0;
	t := KeyWait;
	repeat
    {$IFDEF GUIMSWINMODE}
		w32crt.RedrawScreen;
		kp := w32crt.KeyPressed;
    {$ELSE GUIMSWINMODE}
		kp := KeyPressed;
    {$ENDIF GUIMSWINMODE}

		if not( kp ) then begin
    {$IFDEF GUIMSWINMODE}
			w32crt.Delay( KeysamplingInterval );
    {$ELSE GUIMSWINMODE}
			Delay( KeysamplingInterval );
    {$ENDIF GUIMSWINMODE}
			t := t - KeysamplingInterval;
			a := RPK_TimeEvent;
		end else begin
			a := RPGKey;
			t := 0;
		end;
	until ( t <= 0 );
	RPGKey_nonblocking := a;
end;
  {$ENDIF ASCII_SDL}
{$ENDIF PATCH_HEART}

Function DirKey: Integer;
	{ Get a direction selection from the user. If a standard direction }
	{ key was selected, return its direction (0 is East, increase }
	{ clockwise). See Locale.pp for details. }
	{ Return -1 if no good direction was chosen. }
var
	K: Char;
begin
	K := RPGKey;
{$IFDEF PATCH_GH}
	if (KBD_NONE = K) then begin
	end else
{$ENDIF PATCH_GH}
	if K = KeyMap[ KMC_East ].KCode then begin
		DirKey := 0;
	end else if K = KeyMap[ KMC_SouthEast ].KCode then begin
		DirKey := 1;
	end else if K = KeyMap[ KMC_South ].KCode then begin
		DirKey := 2;
	end else if K = KeyMap[ KMC_SouthWest ].KCode then begin
		DirKey := 3;
	end else if K = KeyMap[ KMC_West ].KCode then begin
		DirKey := 4;
	end else if K = KeyMap[ KMC_NorthWest ].KCode then begin
		DirKey := 5;
	end else if K = KeyMap[ KMC_North ].KCode then begin
		DirKey := 6;
	end else if K = KeyMap[ KMC_NorthEast ].KCode then begin
		DirKey := 7;
{$IFDEF PATCH_GH}
	end else if K = KeyMap[ KMC_Enter ].KCode then begin
		DirKey := 8;
	end else if K = KeyMap[ KMC_Enter2 ].KCode then begin
		DirKey := 8;
	end else if K = ' ' then begin
		DirKey := 8;
	end else if K = #27 then begin
		DirKey := -1;
	end else if K = KeyMap[ KMC_QuitGame ].KCode then begin
		DirKey := -1;
	end else if K = KeyMap[ KMC_Eject ].KCode then begin
		DirKey := -1;
	end else begin
		DirKey := 8;
{$ELSE PATCH_GH}
	end else begin
		DirKey := -1;
{$ENDIF PATCH_GH}
	end;

end;

Procedure EndOfGameMoreKey;
	{ The end of the game has been reached. Wait for the user to }
	{ press either the space bar or the ESC key. }
var
	A: Char;
begin
{$IFDEF PATCH_GH}
  {$IFDEF PATCH_I18N}
	DialogMSG( I18N_MsgString('EndOfGameMoreKey','Hit space bar') );
  {$ELSE PATCH_I18N}
	DialogMSG( '[Hit space bar.]' );
  {$ENDIF PATCH_I18N}
{$ENDIF PATCH_GH}

	{ First, get rid of any pending keypresses. }
{$IF DEFINED(ASCII_SDL)}
	DoFlip;
{$ELSEIF DEFINED(GUIMSWINMODE)}
	while w32crt.keypressed do w32crt.readkey;
{$ELSE}
	while keypressed do readkey;
{$ENDIF}

	{ Keep reading keypresses until either a space or an ESC is found. }
	repeat
		A := RPGKey;
	until ( A = ' ' ) or ( A = #27 );
end;

{$IFDEF ASCII_SDL}
Procedure CMessage( const msg: String; Z: Integer; C: TSDL_Color );
{$ELSE ASCII_SDL}
Procedure CMessage( const msg: String; Z: Integer; C: Byte );
{$ENDIF ASCII_SDL}
	{ Display MSG centered in zone Z. }
var
	X,Y: Integer;
begin
	{ Figure out the coordinates for centered display. }
{$IFDEF PATCH_I18N}
	X := ( ScreenZone[Z,3] + ScreenZone[Z,1] ) div 2 - ( WidthMBCharStr( msg ) div 2 ) + 1;
{$ELSE PATCH_I18N}
	X := ( ScreenZone[Z,3] + ScreenZone[Z,1] ) div 2 - ( Length( msg ) div 2 ) + 1;
{$ENDIF PATCH_I18N}
	Y := ( ScreenZone[Z,4] + ScreenZone[Z,2] ) div 2;

	{ Actually do the output. }
	ClrZone( Z );
	if X < 1 then X := 1;
	if Y < 1 then Y := 1;
{$IF DEFINED(ASCII_SDL)}
	ascii_sdl.GotoXY( X , Y );
	ascii_sdl.TextColor( C );
{$ELSEIF DEFINED(GUIMSWINMODE)}
	w32crt.GotoXY( X , Y );
	w32crt.TextColor( C );
{$ELSE}
	GotoXY( X , Y );
	TextColor( C );
{$ENDIF}
{$IFDEF PATCH_I18N}
	WriteMBCharStr(msg,0);
{$ELSE PATCH_I18N}
	Write(msg);
{$ENDIF PATCH_I18N}
end;

{$IFDEF ASCII_SDL}
Procedure GameMSG( msg: string; X1,Y1,X2,Y2: Byte; C: TSDL_Color );  {not const-able}
{$ELSE ASCII_SDL}
Procedure GameMSG( msg: string; X1,Y1,X2,Y2,C: Byte );  {not const-able}
{$ENDIF ASCII_SDL}
	{Prettyprint the string MSG with color C in screen zone Z.}
var
	Width: Integer;		{Our pixel width.}
	NextWord: String;
	THELine: String;	{The line under construction.}
	LC: Boolean;		{Loop Condition.}
{$IFDEF PATCH_I18N}
	LW_I18N: Boolean;	{Is the last word I18N character?}
	CW_I18N: Boolean;	{Is the current word I18N character?}
	DItS: Boolean;		{Do insert the space, or not.}
{$ENDIF PATCH_I18N}
begin
	{ CLean up the message a bit. }
	DeleteWhiteSpace( msg );
{$IF DEFINED(ASCII_SDL)}
	ascii_sdl.TextColor( C );
	ascii_sdl.TextBackground( StdBlack );
{$ELSEIF DEFINED(GUIMSWINMODE)}
	w32crt.TextColor( C );
	w32crt.TextBackground( StdBlack );
{$ELSE}
	TextColor( C );
	TextBackground( StdBlack );
{$ENDIF}

	{Clear the message area, and set clipping bounds.}
{$IF DEFINED(ASCII_SDL)}
	ascii_sdl.Window( X1 , Y1 , X2 , Y2 );
	ascii_sdl.ClrScr;
{$ELSEIF DEFINED(GUIMSWINMODE)}
	w32crt.Window( X1 , Y1 , X2 , Y2 );
	w32crt.ClrScr;
{$ELSE}
	Window( X1 , Y1 , X2 , Y2 );
	ClrScr;
{$ENDIF}

	{Calculate the width of the text area.}
	Width := X2 - X1;

	{THELine = The first word in this iteration}
{$IFDEF PATCH_I18N}
	LW_I18N := False;
	THELine := ExtractWordForPrint( msg, DItS, CW_I18N );
{$ELSE PATCH_I18N}
	THELine := ExtractWord( msg );
{$ENDIF PATCH_I18N}

	{Start the main processing loop.}
	while TheLine <> '' do begin
		{Set the LoopCondition to True.}
		LC := True;

		{ Start building the line. }
		repeat
{$IFDEF PATCH_I18N}
			NextWord := ExtractWordForPrint( Msg, DItS, CW_I18N );

			if (False = LW_I18N) and (False = CW_I18N) then begin
				DItS := True;
			end;
			LW_I18N := CW_I18N;
			if DItS then begin
				if WidthMBCharStr(THEline + ' ') <= Width then begin
					THEline := THEline + ' ';
				end;
			end;
			if WidthMBCharStr(THEline + NextWord) <= Width then begin
				THEline := THEline + NextWord;
			end else begin
				LC := False;
			end;
{$ELSE PATCH_I18N}
			NextWord := ExtractWord( Msg );

			if Length(THEline + ' ' + NextWord) < Width then
				THEline := THEline + ' ' + NextWord
			else
				LC := False;
{$ENDIF PATCH_I18N}

		until (not LC) or (NextWord = '') or ( TheLine[Length(TheLine)] = #13 );

		{ If the line ended due to a line break, deal with it. }
		if ( TheLine[Length(TheLine)] = #13 ) then begin
			{ Display the line break as a space. }
			TheLine[Length(TheLine)] := ' ';
{$IFDEF PATCH_I18N}
			NextWord := ExtractWordForPrint( msg, DItS, CW_I18N );
{$ELSE PATCH_I18N}
			NextWord := ExtractWord( msg );
{$ENDIF PATCH_I18N}
		end;

		{ Output the line. }
		if NextWord = '' then begin
{$IFDEF PATCH_I18N}
			WriteMBCharStr(THELine,Width);
{$ELSE PATCH_I18N}
			Write(THELine);
{$ENDIF PATCH_I18N}
		end else begin
{$IFDEF PATCH_I18N}
			WriteMBCharStr(THELine,Width);
  {$IF DEFINED(ASCII_SDL)}
			ascii_sdl.WriteLn;
  {$ELSEIF DEFINED(GUIMSWINMODE)}
			w32crt.WriteLn;
  {$ELSE}
			WriteLn;
  {$ENDIF}
{$ELSE PATCH_I18N}
			WriteLn(THELine);
{$ENDIF PATCH_I18N}
		end;

		{ Prepare for the next iteration. }
		TheLine := NextWord;

	end; { while msg <> '' }

	{Restore the clip window to its maximum size.}
	MaxClipZone;
end;

{$IFDEF ASCII_SDL}
Procedure GameMSG( const msg: string; Z: Byte; C: TSDL_Color );
{$ELSE ASCII_SDL}
Procedure GameMSG( const msg: string; Z,C: Byte );
{$ENDIF ASCII_SDL}
	{ Print a message in zone Z. }
begin
	GameMSG( msg , ScreenZone[Z,1], ScreenZone[Z,2], ScreenZone[Z,3], ScreenZone[Z,4], C );
end;

{$IFDEF PATCH_GH}
Procedure RedrawConsole;
	{ Redraw the console. }
var
	CH_tmp: SAttPtr;
	SL: SAttPtr;
	msg: String;
	NextWord: String;
	TheLine: String;
	MaxWidth: Integer;
	MaxHeight: Integer;
begin
	{ Restore the console display. }

	{Clear the message area, and set clipping bounds.}
	ClrZone( ZONE_Dialog );
	ClipZone( ZONE_Dialog );

	{Set initial cursor position.}
  {$IF DEFINED(ASCII_SDL)}
	ascii_sdl.GotoXY( 1 , 1 );
	ascii_sdl.TextColor( InfoGreen );
	ascii_sdl.TextBackground( StdBlack );
  {$ELSEIF DEFINED(GUIMSWINMODE)}
	w32crt.GotoXY( 1 , 1 );
	w32crt.TextColor( InfoGreen );
	w32crt.TextBackground( StdBlack );
  {$ELSE}
	GotoXY( 1 , 1 );
	TextColor( InfoGreen );
	TextBackground( StdBlack );
  {$ENDIF}

	MaxWidth := ScreenZone[ZONE_Dialog,3] - ScreenZone[ZONE_Dialog,1];
	MaxHeight := ScreenZone[ZONE_Dialog,4] - ScreenZone[ZONE_Dialog,2];

	CH_tmp := NIL;

	SL := RetrieveSAtt( Console_History , NumSAtts( Console_History ) - MaxHeight );
	if ( NIL = SL ) then begin
		SL := Console_History;
	end;
	while ( NIL <> SL ) do begin
		msg := SL^.Info;
		TheLine := '';
		NextWord := '';
		repeat
			GetNextLine( TheLine , msg , NextWord , MaxWidth );
			if ( '' <> TheLine ) then begin
				StoreSAtt( CH_tmp , TheLine );
			end;
			TheLine := NextWord;
		until ( '' = TheLine );
		SL := SL^.Next;
	end;

	SL := RetrieveSAtt( CH_tmp , NumSAtts( CH_tmp ) - MaxHeight );
	if ( NIL = SL ) then SL := CH_tmp;
	while ( NIL <> SL ) do begin
  {$IFDEF PATCH_I18N}
		WriteMBCharStr( SL^.Info, MaxWidth );
  {$ELSE PATCH_I18N}
		write( SL^.Info );
  {$ENDIF PATCH_I18N}
  {$IF DEFINED(ASCII_SDL)}
		ascii_sdl.writeln;
  {$ELSEIF DEFINED(GUIMSWINMODE)}
		w32crt.WriteLn;
  {$ELSE}
		writeln;
  {$ENDIF}
		SL := SL^.Next;
	end;

	DisposeSAtt( CH_tmp );

	{Restore the clip window to its maximum size.}
	MaxClipZone;
end;
{$ENDIF PATCH_GH}

Procedure DialogMSG(msg: string); {not const-able}
	{ Print a message in the scrolling dialog box. }
var
{$IFDEF PATCH_GH}
	SA: SAttPtr;
{$ELSE PATCH_GH}
	Width: Integer;		{Our pixel width.}
	NextWord: String;
	THELine: String;	{The line under construction.}
	LC: Boolean;		{Loop Condition.}
	SA: SAttPtr;
{$IFDEF PATCH_I18N}
	LW_I18N: Boolean;	{Is the last word I18N ?}
	CW_I18N: Boolean;	{Is the current word I18N ?}
	DItS: Boolean;		{Do insert the space, or not.}
	SL: SAttPtr;
{$ENDIF PATCH_I18N}
{$ENDIF PATCH_GH}
begin
{$IFDEF DEBUG}
	if DEBUG_TraceMacro then begin
		ErrorMessage_fork( 'TRACE: DialogMSG() "' + msg + '"');
	end;
{$ENDIF DEBUG}
	{ CLean up the message a bit. }
	DeleteWhiteSpace( msg );
{$IF DEFINED(ASCII_SDL)}
	ascii_sdl.TextColor( InfoGreen );
	ascii_sdl.TextBackground( StdBlack );
{$ELSEIF DEFINED(GUIMSWINMODE)}
	w32crt.TextColor( InfoGreen );
	w32crt.TextBackground( StdBlack );
{$ELSE}
	TextColor( InfoGreen );
	TextBackground( StdBlack );
{$ENDIF}
	msg := '> ' + msg;

{$IFDEF PATCH_GH}
	if ( Console_History_Length <= NumSAtts( Console_History ) ) then begin
		SA := Console_History;
		RemoveSAtt( Console_History , SA );
		PurgeSAtt( Console_History );
	end;
	StoreSAtt( Console_History , msg );

	{ Redraw the dialog area. }
	RedrawConsole;
{$ELSE PATCH_GH}
	{Clear the message area, and set clipping bounds.}
	ClipZone( ZONE_Dialog );

	{Set initial cursor position.}
{$IF DEFINED(ASCII_SDL)}
	ascii_sdl.GotoXY( 1 , ScreenZone[ZONE_Dialog,4] - ScreenZone[ZONE_Dialog,2] + 1 );
{$ELSEIF DEFINED(GUIMSWINMODE)}
	w32crt.GotoXY( 1 , ScreenZone[ZONE_Dialog,4] - ScreenZone[ZONE_Dialog,2] + 1 );
{$ELSE}
	GotoXY( 1 , ScreenZone[ZONE_Dialog,4] - ScreenZone[ZONE_Dialog,2] + 1 );
{$ENDIF}

	{Calculate the width of the text area.}
	Width := ScreenZone[ZONE_Dialog,3] - ScreenZone[ZONE_Dialog,1];

	{THELine = The first word in this iteration}
{$IFDEF PATCH_I18N}
	LW_I18N := False;
	THELine := ExtractWordForPrint( msg, DItS, CW_I18N );
{$ELSE PATCH_I18N}
	THELine := ExtractWord( msg );
{$ENDIF PATCH_I18N}

	{Start the main processing loop.}
	while TheLine <> '' do begin
		{Set the LoopCondition to True.}
		LC := True;

		{ Start building the line. }
		repeat
{$IFDEF PATCH_I18N}
			NextWord := ExtractWordForPrint( Msg, DItS, CW_I18N );

			if (False = LW_I18N) and (False = CW_I18N) then begin
				DItS := True;
			end;
			LW_I18N := CW_I18N;
			if DItS then begin
				if WidthMBCharStr(THEline + ' ') <= Width then begin
					THEline := THEline + ' ';
				end;
			end;
			if WidthMBCharStr(THEline + NextWord) <= Width then begin
				THEline := THEline + NextWord;
			end else begin
				LC := False;
			end;
{$ELSE PATCH_I18N}
			NextWord := ExtractWord( Msg );

			if Length(THEline + ' ' + NextWord) < Width then
				THEline := THEline + ' ' + NextWord
			else
				LC := False;
{$ENDIF PATCH_I18N}

		until (not LC) or (NextWord = '') or ( TheLine[Length(TheLine)] = #13 );

		{ If the line ended due to a line break, deal with it. }
		if ( TheLine[Length(TheLine)] = #13 ) then begin
			{ Display the line break as a space. }
			TheLine[Length(TheLine)] := ' ';
{$IFDEF PATCH_I18N}
			NextWord := ExtractWordForPrint( msg, DItS, CW_I18N );
{$ELSE PATCH_I18N}
			NextWord := ExtractWord( msg );
{$ENDIF PATCH_I18N}
		end;

		{ Output the line. }
		if TheLine <> '' then begin
{$IFDEF PATCH_I18N}
{$ELSE PATCH_I18N}
			writeln;
			write( TheLine );
{$ENDIF PATCH_I18N}
			if NumSAtts( Console_History ) >= Console_History_Length then begin
				SA := Console_History;
				RemoveSAtt( Console_History , SA );
{$IFDEF PATCH_GH}
				PurgeSAtt( Console_History );
{$ENDIF PATCH_GH}
			end;
			StoreSAtt( Console_History , TheLine );
		end;

		{ Prepare for the next iteration. }
		TheLine := NextWord;

	end; { while msg <> '' }

{$IFDEF PATCH_I18N}
	{ NOTE: In CJK, there are many charctors, one charactor have double size for one ANK charactor and data length is 3 or 4 bytes. }
	{ NOTE: But, a function Writeln() fail scrolling these charactors. }
	{ NOTE: FPC's unicode functions is not stable, is its? }

	ClrZone( ZONE_Dialog );
	MaxClipZone;

	{ Restore the console display. }
  {$IF DEFINED(ASCII_SDL)}
	ascii_sdl.GotoXY( 1 , 1 );
	ascii_sdl.TextColor( Green );
  {$ELSEIF DEFINED(GUIMSWINMODE)}
	w32crt.GotoXY( 1 , 1 );
	w32crt.TextColor( Green );
  {$ELSE}
	GotoXY( 1 , 1 );
	TextColor( Green );
  {$ENDIF}
	SL := RetrieveSAtt( Console_History , NumSAtts( Console_History ) - ScreenRows + ScreenZone[ ZONE_Dialog , 2 ] );
	if SL = Nil then SL := Console_History;
	while SL <> Nil do begin
  {$IF DEFINED(ASCII_SDL)}
		ascii_sdl.writeln;
  {$ELSEIF DEFINED(GUIMSWINMODE)}
		w32crt.WriteLn;
  {$ELSE}
		writeln;
  {$ENDIF}
		WriteMBCharStr( SL^.Info, Width );
		SL := SL^.Next;
	end;
{$ELSE PATCH_I18N}
	{Restore the clip window to its maximum size.}
	MaxClipZone;
{$ENDIF PATCH_I18N}

{$ENDIF PATCH_GH}
end;

{$IFDEF ASCII_SDL}
{$IFDEF PATCH_GH}
Function GetStringFromUser( const Prompt, Init_text: String; var canceled: Boolean ): String;
begin
	GetStringFromUser := GetStringFromUser_Raw( Prompt , Init_text, canceled );
end;

Function GetStringFromUser( const Prompt, Init_text: String ): String;
var
	dummy: Boolean;
begin
	GetStringFromUser := GetStringFromUser_Raw( Prompt , Init_text, dummy );
end;
{$ENDIF PATCH_GH}
Function GetStringFromUser( const Prompt: String ): String;
begin
	GetStringFromUser := GetStringFromUser_Raw( Prompt );
end;
{$ELSE ASCII_SDL}
{$IFDEF PATCH_GH}
Function GetStringFromUser( const Prompt, Init_text: String ): String;
var
	dummy: Boolean;
begin
	GetStringFromUser := GetStringFromUser( Prompt, Init_text, dummy );
end;

Function GetStringFromUser( const Prompt: String ): String;
var
	dummy: Boolean;
begin
	GetStringFromUser := GetStringFromUser( Prompt, '', dummy );
end;

Function GetStringFromUser( const Prompt, Init_text: String; var canceled: Boolean ): String;
{$ELSE PATCH_GH}
Function GetStringFromUser( const Prompt: String ): String;
{$ENDIF PATCH_GH}
	{ Does what it says. }
{$IFDEF GUIMSWINMODE}
const
	WCLen = 288; { 254; }
{$ENDIF GUIMSWINMODE}
var
	it: String;
{$IFDEF PATCH_I18N}
  {$IFDEF GUIMSWINMODE}
	pmsg: PChar;
	pdst: PChar;
	dstlen: Integer;
	dst: String;
  {$ELSE GUIMSWINMODE}
	RK: Char;
	state: ShortInt = 0;
	mbchar_work: String = '';
  {$ENDIF GUIMSWINMODE}
	MaxInputWidth: Integer = 0;
  {$IFDEF PATCH_GH}
	trimedlength: integer;
	getit: Char;
  {$ENDIF PATCH_GH}
	X: Integer;
{$ELSE PATCH_I18N}
  {$IFDEF PATCH_GH}
	MaxInputWidth: Integer = 0;
	X: Integer;
  {$ENDIF PATCH_GH}
{$ENDIF PATCH_I18N}
begin
{$IFDEF PATCH_GH}
	canceled := False;
	DrawZoneBorder( ScreenZone[ ZONE_TextInputBigBox , 1 ] - 1 , ScreenZone[ ZONE_TextInputBigBox , 2 ] -1 , ScreenZone[ ZONE_TextInputBigBox , 3 ] + 1 , ScreenZone[ ZONE_TextInputBigBox , 4 ] + 1 , LightCyan );
	ClrZone( ZONE_TextInputBigBox );
  {$IFDEF PATCH_I18N}
	GameMSG( Prompt , ZONE_TextInputPrompt , InfoGreen );
  {$ELSE PATCH_I18N}
	X := ( ScreenZone[ZONE_TextInputPrompt,3] + ScreenZone[ZONE_TextInputPrompt,1] ) div 2 - ( WidthMBCharStr( Prompt ) div 2 ) + 1;
	if X < GOTOXY_MIN then begin
		X := GOTOXY_MIN;
	end;
	GotoXY( X, ScreenZone[ ZONE_TextInputPrompt , 2 ] );
	TextColor( InfoGreen );
	Write( Prompt );
  {$ENDIF PATCH_I18N}
{$ELSE PATCH_GH}
	DrawZoneBorder( ScreenZone[ ZONE_TextInput , 1 ] - 1 , ScreenZone[ ZONE_TextInput , 2 ] -1 , ScreenZone[ ZONE_TextInput , 3 ] + 1 , ScreenZone[ ZONE_TextInput , 4 ] + 1 , LightCyan );
	ClrZone( ZONE_TextInput );
  {$IFDEF PATCH_I18N}
	X := ( ScreenZone[ZONE_TextInput,3] + ScreenZone[ZONE_TextInput,1] ) div 2 - ( WidthMBCharStr( Prompt ) div 2 ) + 1;
	if X < GOTOXY_MIN then begin
		X := GOTOXY_MIN;
	end;
    {$IFDEF GUIMSWINMODE}
	w32crt.GotoXY( X, ScreenZone[ ZONE_TextInput , 4 ] );
	w32crt.TextColor( InfoGreen );
	WriteMBCharStr( Prompt, 0 );
    {$ELSE GUIMSWINMODE}
	GotoXY( X, ScreenZone[ ZONE_TextInput , 4 ] );
	TextColor( InfoGreen );
	WriteMBCharStr( Prompt, 0 );
    {$ENDIF GUIMSWINMODE}
  {$ELSE PATCH_I18N}
	GotoXY( ( ScreenZone[ZONE_TextInput,3] + ScreenZone[ZONE_TextInput,1] ) div 2 - ( Length( Prompt ) div 2 ) + 1 , ScreenZone[ ZONE_TextInput , 4 ] );
	TextColor( InfoGreen );
	Write( Prompt );
  {$ENDIF PATCH_I18N}
{$ENDIF PATCH_GH}
	TextColor( InfoHilight );
	CursorOn;
{$IFDEF GUIMSWINMODE}
	w32crt.RedrawScreen;
{$ENDIF GUIMSWINMODE}
	ClipZone( ZONE_TextInput );

{$IFDEF PATCH_GH}
	MaxInputWidth := ScreenZone[ ZONE_TextInput , 3 ] - ScreenZone[ ZONE_TextInput , 1 ] + 1;
	if 127 < MaxInputWidth then MaxInputWidth := 127;

  {$IFDEF PATCH_I18N}
	trimedlength := MBCharTrimedLength( Init_text, MaxInputWidth );
	if (0 < trimedlength) then begin
		it := Copy(Init_text,1,trimedlength);
	end else begin
		it := '';
	end;
  {$ELSE PATCH_I18N}
	it := Copy(Init_text,1,MaxInputWidth);
  {$ENDIF PATCH_I18N}

  {$IFDEF GUIMSWINMODE}
	w32crt.GotoXY( 1 , 1 );
	w32crt.ClrEOL;

	pdst := sysutils.StrAlloc(WCLen);
	pmsg := QuickPCopy(it);
	dstlen := Conv_ToTenc( pmsg, Length(it), pdst, WCLen );
	dst := pdst;
	it := w32crt.GetLine( dst , canceled );
	sysutils.StrDispose( pmsg );
	pmsg := QuickPCopy(it);
	dstlen := Conv_FromTenc( pmsg, Length(it), pdst, WCLen );
	it := pdst;
	sysutils.StrDispose( pmsg );
	sysutils.StrDispose( pdst );
  {$ELSE GUIMSWINMODE}
    {$IFDEF PATCH_I18N}
	GotoXY( 1 , 1 );
	ClrEOL;
	WriteMBCharStr( it, MaxInputWidth );

	repeat
		if TERMINAL_bidiRTL then begin
			GotoXY( 1, 1 );
		end else begin
			GotoXY( 1 + WidthMBCharStr(it) , 1 );
		end;
		repeat
			RK := ReadKey;
{$IFDEF DEBUG}
			WriteLn(IntToHex(Ord(RK),2));
{$ENDIF DEBUG}
			{ 45 5B 1B: 5 in TenKey on Unix-Console }
			if #0 = RK then begin
				{We have a two-part special key.}
				getit := ReadKey;
{$IFDEF DEBUG}
				WriteLn(IntToHex(Ord(getit),2));
{$ENDIF DEBUG}
				case getit of
					#$52: RK := '0';	{ 0 in TenKey }
					#$53: RK := '.';	{ . in TenKey on X-Window/MS-Windows }
					#$5A: RK := '1';	{ End Cursor Key on X-Window }
					#$4F: RK := '1';	{ End Cursor Key on Unix-Console, MS-Windows }
					#$50: RK := '2';	{ Down Cursor Key }
					#$51: RK := '3';	{ PageDown Cursor Key }
					#$4B: RK := '4';	{ Left Cursor Key }
					#$5F: RK := '5';	{ 5 in TenKey on X-Window }
					#$4C: RK := '5';	{ 5 in TenKey on MS-Windows }
					#$4D: RK := '6';	{ Right Cursor Key }
					#$5B: begin
							getit := ReadKey;
							case getit of
								#$1B: RK := '7';	{ Home Cursor Key on X-Window }
								else RK := #0;
							end;
						end;
					#$47: RK := '7';	{ Home Cursor Key on Unix-Console, MS-Windows }
					#$48: RK := '8';	{ Up Cursor Key }
					#$49: RK := '9';	{ PageUp Cursor Key }
					#$35: RK := '/';	{ Slash in TenKey on MS-Windows }
					#$1C: RK := #$0A;	{ Enter in TenKey on MS-Windows => Altanative-RET }
					{ #$3B: }{ F1  on Unix-Console, MS-Windows }
					{ #$3C: }{ F2  on Unix-Console, MS-Windows }
					{ #$3D: }{ F3  on Unix-Console, MS-Windows }
					{ #$3E: }{ F4  on Unix-Console, MS-Windows }
					{ #$3F: }{ F5  on Unix-Console, MS-Windows }
					{ #$40: }{ F6  }
					{ #$41: }{ F7  }
					{ #$42: }{ F8  }
					{ #$43: }{ F9  }
					{ #$44: }{ F10 }
					{ #$85: }{ F11 }
					{ #$86: }{ F12 }
					#$29: RK := #$1B;	{ JP109key-Hankaku/Zenkaku on MS-Windows => ESC }
					else RK := #0;
				end;
			end;
			RK := EditMBCharStr( it, 127, MaxInputWidth, RK, NIL, state, mbchar_work );
		until not(RK = #255);
		GotoXY( 1 , 1 );
		ClrEOL;
		WriteMBCharStr( it, MaxInputWidth );
	until (RK = #10) or (RK = #13) or (RK = #27);
	if (#27 = RK) then begin
		it := '';
		canceled := True;
	end;
    {$ELSE PATCH_I18N}
	GotoXY( 1 , 1 );
	ClrEOL;
	Write( it );
	{ ***BUG*** }
	GotoXY( 1 , 1 );
	ReadLn( it );
    {$ENDIF PATCH_I18N}
  {$ENDIF GUIMSWINMODE}
{$ELSE PATCH_GH}
  {$IFDEF PATCH_I18N}
    {$IFDEF GUIMSWINMODE}
	w32crt.GotoXY( 1 , 1 );
	it := w32crt.GetLine;
    {$ELSE GUIMSWINMODE}
	it := '';
	MaxInputWidth := ScreenZone[ ZONE_TextInput , 3 ] - ScreenZone[ ZONE_TextInput , 1 ] + 1;
	if 127 < MaxInputWidth then MaxInputWidth := 127;
	repeat
		if TERMINAL_bidiRTL then begin
			GotoXY( 1, 1 );
		end else begin
			GotoXY( 1 + WidthMBCharStr(it) , 1 );
		end;
		repeat
			RK := ReadKey;
			RK := EditMBCharStr( it, 127, MaxInputWidth, RK, NIL, state, mbchar_work );
		until not(RK = #255);
		GotoXY( 1 , 1 );
		ClrEOL;
		WriteMBCharStr( it, MaxInputWidth );
	until (RK = #10) or (RK = #13) or (RK = #27);
    {$ENDIF GUIMSWINMODE}
  {$ELSE PATCH_I18N}
	GotoXY( 1 , 1 );
	ReadLn( it );
  {$ENDIF PATCH_I18N}
{$ENDIF PATCH_GH}

{$IFDEF GUIMSWINMODE}
	w32crt.CursorOff;
{$ELSE GUIMSWINMODE}
	CursorOff;
{$ENDIF GUIMSWINMODE}
	ClrZone( ZONE_Map );
	MaxClipZone;

	GetStringFromUser := it;
end;
{$ENDIF ASCII_SDL}

Function MsgString( const MsgLabel: String ): String;
	{ Return the standard message string which has the requested }
	{ label. }
begin
	MsgString := SAttValue( Text_Messages , MsgLabel );
end;

{$IFDEF PATCH_GH}
{$ELSE PATCH_GH}
Function MoreHighFirstLine( LList: SAttPtr ): Integer;
	{ Determine the highest possible FirstLine value. }
var
	it: Integer;
begin
{$IFDEF PATCH_GH}
	if (NIL = LList) then Exit(0);
{$ENDIF PATCH_GH}
	it := NumSAtts( LList ) - ( ScreenRows - 3 );
	if it < 1 then it := 1;
	MoreHighFirstLine := it;
end;
{$ENDIF PATCH_GH}

Procedure MoreText( LList: SAttPtr; FirstLine: Integer );
	{ Browse this text file across the majority of the screen. }
	{ Clear the screen upon exiting, though restoration of the }
	{ previous display is someone else's responsibility. }
{$IFDEF PATCH_GH}
var
	MT_tmp: SAttPtr;
{$ENDIF PATCH_GH}

	Procedure DisplayTextHere;
	var
		CLine: SAttPtr;	{ Current Line }
	{$IFDEF PATCH_I18N}
		trimedlength: integer;
	{$ENDIF}
	begin
		{ Error check. }
	{$IFDEF PATCH_GH}
		if ( -1 = FirstLine ) then begin
			FirstLine := NumSAtts( MT_tmp ) - ( ScreenRows - 3 );
		end;
		if ( FirstLine < 1 ) then begin
			FirstLine := 1
		end else if ( NumSAtts( MT_tmp ) < FirstLine ) then begin
			FirstLine := NumSAtts( MT_tmp );
		end;
	{$ELSE PATCH_GH}
		if FirstLine < 1 then FirstLine := 1
		else if FirstLine > MoreHighFirstLine( LList ) then FirstLine := MoreHighFirstLine( LList );
	{$ENDIF PATCH_GH}
	{$IF DEFINED(ASCII_SDL)}
		ascii_sdl.GotoXY( 1 , 1 );
	{$ELSEIF DEFINED(GUIMSWINMODE)}
		w32crt.GotoXY( 1 , 1 );
	{$ELSE}
		GotoXY( 1 , 1 );
	{$ENDIF}

	{$IFDEF PATCH_GH}
		CLine := RetrieveSATt( MT_tmp , FirstLine );
	{$ELSE PATCH_GH}
		CLine := RetrieveSATt( LList , FirstLine );
	{$ENDIF PATCH_GH}
	{$IF DEFINED(ASCII_SDL)}
		while ( ascii_sdl.WhereY < ( ScreenRows - 1 ) ) do begin
			ascii_sdl.ClrEOL;
	{$ELSEIF DEFINED(GUIMSWINMODE)}
		while ( w32crt.WhereY < ( ScreenRows - 1 ) ) do begin
			w32crt.ClrEOL;
	{$ELSE}
		while ( WhereY < ( ScreenRows - 1 ) ) do begin
			ClrEOL;
	{$ENDIF}
			if CLine <> Nil then begin
	{$IFDEF PATCH_I18N}
				trimedlength := MBCharTrimedLength( CLine^.Info, ScreenColumns - 2 );
				if (0 < trimedlength) then begin
					WriteMBCharStr( Copy(CLine^.Info,1,trimedlength), ScreenColumns );
				end;
	  {$IF DEFINED(ASCII_SDL)}
				ascii_sdl.WriteLn;
	  {$ELSEIF DEFINED(GUIMSWINMODE)}
				w32crt.WriteLn;
	  {$ELSE}
				WriteLn;
	  {$ENDIF}
	{$ELSE PATCH_I18N}
				writeln( Copy( CLine^.Info , 1 , ScreenColumns - 2 ) );
	{$ENDIF PATCH_I18N}
				CLine := CLine^.Next;
			end else begin
	{$IF DEFINED(ASCII_SDL)}
				ascii_sdl.writeln;
	{$ELSEIF DEFINED(GUIMSWINMODE)}
				w32crt.WriteLn;
	{$ELSE}
				writeln;
	{$ENDIF}
			end;
		end;
{$IFDEF ASCII_SDL}
		DoFlip;
{$ENDIF ASCII_SDL}
	end;
var
	A: Char;
{$IFDEF PATCH_GH}
	MaxWidth: Integer;
	MaxHeight: Integer;
	RPM: RPGMenuPtr;
	t, t_max: Integer;
	CLine: SAttPtr;	{ Current Line }
	msg: String;
	NextWord: String;
	TheLine: String;
	cont: Boolean;
	stored: Boolean;
{$ENDIF PATCH_GH}
begin
{$IF DEFINED(ASCII_SDL)}
	ascii_sdl.ClrScr;
	ascii_sdl.GotoXY( 1 , ScreenROws );
	ascii_sdl.TextColor( LightGreen );
	ascii_sdl.TextBackground( Black );
{$ELSEIF DEFINED(GUIMSWINMODE)}
	w32crt.ClrScr;
	w32crt.GotoXY( 1 , ScreenRows );
	w32crt.TextColor( LightGreen );
	w32crt.TextBackground( Black );
{$ELSE}
	ClrScr;
	GotoXY( 1 , ScreenROws );
	TextColor( LightGreen );
	TextBackground( Black );
{$ENDIF}
{$IFDEF PATCH_I18N}
	WriteMBCharStr( MsgString( 'MORETEXT_Prompt' ), ScreenColumns );
{$ELSE PATCH_I18N}
	Write( MsgString( 'MORETEXT_Prompt' ) );
{$ENDIF PATCH_I18N}

{$IFDEF PATCH_GH}
	MaxWidth := ScreenZone[ZONE_MoreText,3] - ScreenZone[ZONE_MoreText,1] + 1;
	MaxHeight := ScreenZone[ZONE_MoreText,4] - ScreenZone[ZONE_MoreText,2] + 1;
	if Show_MenuScrollbar then begin
		RPM := CreateRPGMenu( LightGray , Green , ZONE_MoreText );
		CLine := RetrieveSAtt( LList , 1 );
		t_max := NumSAtts( LList );
		for t := 1 to t_max do begin
			msg := CLine^.Info;
			TheLine := '';
			NextWord := '';
			cont := false;
			stored := false;
			repeat
				GetNextLine( TheLine , msg , NextWord , MaxWidth );
				if ('' <> TheLine) then begin
					if cont then begin
						AddRPGMenuItem( RPM , #$0 + TheLine , -1 );
					end else begin
						AddRPGMenuItem( RPM , TheLine , -1 );
						cont := true;
					end;
					stored := true;
				end;
				TheLine := NextWord;
			until ('' = TheLine);
			if not stored then begin
				AddRPGMenuItem( RPM , '' , -1 );
			end;
			CLine := CLine^.Next;
		end;
		if ( -1 = FirstLine ) then begin
			FirstLine := RPM^.NumItem - MaxHeight;
		end;
		if ( FirstLine < 1 ) then begin
			FirstLine := 1
		end else if ( RPM^.NumItem < FirstLine ) then begin
			FirstLine := RPM^.NumItem;
		end;
		RPM^.TopItem := FirstLine;
		SetItemByPosition( RPM, FirstLine );
		SelectMenu( RPM );
		DisposeRPGMenu( RPM );
	end else begin
{$ENDIF PATCH_GH}
{$IFDEF PATCH_GH}
	MT_tmp := NIL;
	CLine := LList;
	while ( NIL <> CLine ) do begin
		msg := CLine^.Info;
		TheLine := '';
		NextWord := '';
		stored := false;
		repeat
			GetNextLine( TheLine , msg , NextWord , MaxWidth );
			if ( '' <> TheLine ) then begin
				StoreSAtt( MT_tmp , TheLine );
				stored := true;
			end;
			TheLine := NextWord;
		until ( '' = TheLine );
		if not stored then begin
			StoreSAtt( MT_tmp , '' );
		end;
		CLine := CLine^.Next;
	end;
{$ENDIF PATCH_GH}

	{ Display the screen. }
{$IF DEFINED(ASCII_SDL)}
	ascii_sdl.TextColor( LightGray );
{$ELSEIF DEFINED(GUIMSWINMODE)}
	w32crt.TextColor( LightGray );
{$ELSE}
	TextColor( LightGray );
{$ENDIF}
	DisplayTextHere;

	repeat
{$IF DEFINED(ASCII_SDL)}
		DoFlip;
{$ELSEIF DEFINED(GUIMSWINMODE)}
		w32crt.RedrawScreen;
{$ENDIF}
		{ Get input from user. }
		A := RPGKey;

{$IFDEF PATCH_GH}
		{ Possibly process this input. }
		if (KBD_NONE = A) then begin
		end else if (KeyMap[ KMC_MenuUp    ].KCode = A) then begin Dec( FirstLine ); DisplayTextHere;
		end else if (KeyMap[ KMC_MenuDown  ].KCode = A) then begin Inc( FirstLine ); DisplayTextHere;
		end else if (KeyMap[ KMC_PageUp    ].KCode = A) then begin FirstLine := FirstLine - (ScreenRows - 3); DisplayTextHere;
		end else if (KeyMap[ KMC_PageDown  ].KCode = A) then begin FirstLine := FirstLine + (ScreenRows - 3); DisplayTextHere;
		end else if (KeyMap[ KMC_ScrollUp  ].KCode = A) then begin FirstLine := FirstLine - (ScreenRows - 3); DisplayTextHere;
		end else if (KeyMap[ KMC_ScrollDown].KCode = A) then begin FirstLine := FirstLine + (ScreenRows - 3); DisplayTextHere;
		end;

	until (KBD_ESC = A) or ('Q' = A);
{$ELSE PATCH_GH}
		{ Possibly process this input. }
		if A = KeyMap[ KMC_South ].KCode then begin
			Inc( FirstLine );
			DisplayTextHere;
		end else if A = KeyMap[ KMC_North ].KCode then begin
			Dec( FirstLine );
			DisplayTextHere;
		end;

	until ( A = #27 ) or ( A = 'Q' );
{$ENDIF PATCH_GH}
{$IFDEF PATCH_GH}
	DisposeSAtt( MT_tmp );
{$ENDIF PATCH_GH}
{$IFDEF PATCH_GH}
	end;
{$ENDIF PATCH_GH}

	{ CLear the display area. }
{$IF DEFINED(ASCII_SDL)}
	ascii_sdl.ClrScr;
{$ELSEIF DEFINED(GUIMSWINMODE)}
	w32crt.ClrScr;
{$ELSE}
	ClrScr;
{$ENDIF}
end;

{$IFDEF PATCH_GH}
Procedure SetupHQDisplay;
	{ CLear the screen & draw boxes. }
begin
  {$IF DEFINED(ASCII_SDL)}
	ascii_sdl.ClrScr;
  {$ELSEIF DEFINED(GUIMSWINMODE)}
	w32crt.ClrScr;
  {$ELSE}
	ClrScr;
  {$ENDIF}
	RedrawConsole;
end;
{$ENDIF PATCH_GH}



initialization
begin
{$IFDEF DEBUG}
	ErrorMessage_fork('DEBUG: context.pp');
{$ENDIF DEBUG}
	Text_Messages := LoadStringList( Standard_Message_File );
	Console_History := Nil;
{$IFDEF PATCH_GH}
	Attach_SmartPointer( 'Console_History: SAttPtr', @Console_History );
{$ENDIF PATCH_GH}
end;

finalization
begin
{$IFDEF DEBUG}
	ErrorMessage_fork('DEBUG: context.pp(finalization)');
{$ENDIF DEBUG}
	DisposeSAtt( Text_Messages );
	DisposeSAtt( Console_History );
end;

end.
